/*
 * libkarma/karma.c
 *
 * Copyright (c) Frank Zschockelt <libkarma@freakysoft.de> 2004
 *
 * You may distribute and modify this program under the terms of 
 * the GNU GPL, version 2 or later.
 *
 */

#include <sys/socket.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <endian.h>

#include "karma.h"
#include "errors.h"
#include "properties.h"
#include "md5.h"

static int busy_handler(int rio, uint32_t waitfor);

static uint32_t recv_header(int rio);
static int32_t recv_status(int rio);
static int32_t recv_properties(int rio, uint32_t identifier, 
			       uint32_t r_id, char ** properties);

/* 
  Help-Functions
*/

int connectkarma(char * ip)
{
	struct hostent *host;
	struct sockaddr_in dest;
	int rio;
	
	host = gethostbyname(ip);
	if(host == NULL){
		lkerrSet(E_HOST);
		return -1; /*err: gethostbyname() failed (see gethostbyname)*/
	}

	rio = socket(AF_INET, SOCK_STREAM, 0);
	if(rio == -1){
		lkerrSet(E_SOCKET);
		return -1; /* err: socket() failed (see socket)*/
	}
	
	memset(&dest, 0, sizeof(dest));
	dest.sin_addr = *(struct in_addr *)host->h_addr_list[0];
	dest.sin_family = AF_INET;
	dest.sin_port = htons(RIOPORT);

	if(connect(rio, (struct sockaddr *)&dest, sizeof(dest)) == -1){
		lkerrSet(E_CONNECT);
		return -1; /*err: connect() failed (see connect)*/
	} else return rio;	
}

/*
( ((x) & 0xff000000) >> 24) 
| ((x) & 0x00ff0000) >>  8) 
| ((x) & 0x0000ff00) <<  8) 
| ((x) & 0x000000ff) << 24)
*/
uint32_t htorl(uint32_t hostlong)
{
#if __BYTE_ORDER==__BIG_ENDIAN
	return (hostlong>>24) | 
		((hostlong&0xff0000)>>8) | 
		((hostlong&0xff00)<<8) | 
		(hostlong<<24);
#else
	return hostlong;
#endif
}

uint32_t rtohl(uint32_t hostlong)
{
	return htorl(hostlong);
}


/*
((((x) & 0xff00000000000000ull) >> 56)
| (((x) & 0x00ff000000000000ull) >> 40)                             
| (((x) & 0x0000ff0000000000ull) >> 24)                                
| (((x) & 0x000000ff00000000ull) >> 8)                                 
| (((x) & 0x00000000ff000000ull) << 8)                                
| (((x) & 0x0000000000ff0000ull) << 24)                              
| (((x) & 0x000000000000ff00ull) << 40)                              
| (((x) & 0x00000000000000ffull) << 56))
*/
uint64_t htorll(uint64_t hostlong)
{
#if __BYTE_ORDER==__BIG_ENDIAN
	return (hostlong>>56) | 
	 	((hostlong&0xff000000000000)>>40) |
		((hostlong&0xff0000000000)>>24) |
		((hostlong&0xff00000000)>>8) |
		((hostlong&0xff000000)<<8) |
		((hostlong&0xff0000)<<24) |
	 	((hostlong&0xff00)<<40) |
		(hostlong<<56);
#else
	return hostlong;
#endif
}
		
uint64_t rtohll(uint64_t hostlong)
{
	return htorll(hostlong);
}

/*
	 {thanks Cliff L. Biffle...}
	 computes the RID of a file
	 rid algorithm pseudocode:
	 if(length of file <= 64k) 
	 	return MD5(file);
	 else
	 	A = MD5(first 64k);
	 	B = MD5(last 64k);
	 	C = MD5(64k starting at (length - 64k)/2);
	 	return A ^ B ^ C; 
	 
	 oggs,flacs: offset=0 length=length of file
	 mp3s: offset=after the tags length=length of file -128
*/
#define RID_BLOCKS 65536
unsigned char * generate_rid(int file, int offset, int length)
{
	static unsigned char rid_output[32];
	int i;
	uint32_t size;
	unsigned char cache[RID_BLOCKS];
	unsigned char md5output[16], b_md5[16], c_md5[16];
	unsigned char *ret;
	md5_context ctx;
	
	if(lseek(file, offset, SEEK_SET)==-1) return NULL;
	size=read(file, &cache, RID_BLOCKS);
	
	md5_starts(&ctx);
	if(length-offset < RID_BLOCKS){
		md5_update(&ctx, (uint8 *)&cache, size);
		md5_finish(&ctx, md5output);
	}else{
		md5_update(&ctx, cache, RID_BLOCKS);
		md5_finish(&ctx, md5output);
		lseek(file, length-RID_BLOCKS, SEEK_SET);
		read(file, &cache, RID_BLOCKS);
		md5_starts(&ctx);
		md5_update(&ctx, cache, RID_BLOCKS);
		md5_finish(&ctx, b_md5);
		lseek(file, (offset + length - RID_BLOCKS)/2, SEEK_SET);
		size=read(file, &cache, RID_BLOCKS);
		md5_starts(&ctx);
		md5_update(&ctx, cache, size);
		md5_finish(&ctx, c_md5);
		for(i=0; i < 16; i++)
			md5output[i]=md5output[i]^b_md5[i]^c_md5[i];
	}
	for(i=0; i<16;i++){
		if((md5output[i]>>4)<10) rid_output[i*2]=(md5output[i]>>4)+'0';
		else rid_output[i*2]=(md5output[i]>>4)+87;
		if((md5output[i]%16)<10) rid_output[i*2+1]=(md5output[i]%16)+'0';
		else rid_output[i*2+1]=(md5output[i]%16)+87;
	}
	rid_output[32]='\0';
	ret=&rid_output[0];
	return ret;
}


int send_request(int rio, uint32_t identifier, char * payload, int plen)
{
	char msg[HEADERLENGTH];

	/*protocol signature*/
	memcpy(msg, RIO_HEADER, 4);
	/*command number from identifier*/
	identifier=htorl(identifier);
	memcpy(&msg[4], &identifier, sizeof(identifier));

	if(payload==NULL)
		send(rio, msg, HEADERLENGTH, MSG_FIN);
	else{
		send(rio, msg, HEADERLENGTH, 0);
		send(rio, payload, plen, MSG_FIN);
	}
	
	return 0;
}

/*
	 recv the header of the karma,
	 it'll return -1 if the first part of the header
	 doesn't match with the standard rio header, but if it
	 does, it'll return the command identifier
*/
static uint32_t recv_header(int rio)
{
	uint32_t h;
	char sig[4];
	
	recv(rio, &sig, 4, MSG_WAITALL);
	if(memcmp(sig, RIO_HEADER, 4)){
		lkerrSet(E_BADHEADER);
		return -1; /* err: broken Rio Header */
	}
	recv(rio, &h, sizeof(h), MSG_WAITALL);
	return rtohl(h);
}

/*
	 returns the status as int32_t
*/
static int32_t recv_status(int rio)
{
	int32_t s;
	recv(rio, &s, sizeof(s), MSG_WAITALL);
	return rtohl(s);
}

/*
	 recv the properties,
	 copies the data to **properties, 
	 memory will be allocated by recv_properties,
	 returns -1 if the header doesn't match with identifier 
	 or the status indicates an failure.
*/
static int32_t recv_properties(int rio, uint32_t identifier,
			       uint32_t r_id, char ** properties)
{
	uint64_t i=0;
	char id[3];

	if(r_id > 0){
		r_id=htorl(r_id);
		memcpy(id, &r_id, sizeof(r_id));
		send_request(rio, identifier, id, 4);
	}else send_request(rio, identifier, NULL, 0);

	*properties=NULL;
	
	if(busy_handler(rio, identifier)){
		if(recv_status(rio) != 0){
			lkerrSet(E_FAILEDREQ);
			return -1; /*err: failed request*/
		}
	}else return -1; /* already set by busy_handler */

	do{
		if(i%8192==0)
			*properties=(char *)realloc(*properties, (i+8193));
		(*properties)[i+4]='!';
		recv(rio, &(*properties)[i], 4, MSG_WAITALL);
		i+=4;
	}while((*properties)[i-1]!=0);

	return 0;
}

static int busy_handler(int rio, uint32_t waitfor)
{
	uint32_t header;
	header=rtohl(recv_header(rio));
	if(header==BUSY){
		do{
			recv_status(rio);
			recv_status(rio);
		}while((header=rtohl(recv_header(rio)))==BUSY);
	}
	if(header==waitfor)
		return 1;
	else {
		lkerrSet(E_BADHEADER);
		return 0;
	}
}

/* 
  Basic Protocol Functions
*/

int get_protocol_version(int rio, uint32_t * major_version,
			 uint32_t * minor_version)
{
	char versions[8];

	*major_version=htorl(*major_version);
	*minor_version=htorl(*minor_version);
	memcpy(versions, major_version, sizeof(major_version));
	memcpy(&versions[4], minor_version, sizeof(minor_version));

	send_request(rio, 0, versions, 8);
	if(!busy_handler(rio, GET_PROTOCOL_VERSION))
		return -1; /*already set by busy_handler*/

	recv(rio, major_version, sizeof(uint32_t), 0);
	recv(rio, minor_version, sizeof(uint32_t), 0);
	*major_version=rtohl(*major_version);
	*minor_version=rtohl(*minor_version);
	return 0;
}

int32_t get_authentication_salt(int rio, char ** salt)
{
	*salt=(char *)malloc(SALTLENGTH);

	send_request(rio, GET_AUTHENTICATION_SALT, NULL, 0);
	if(!busy_handler(rio, GET_AUTHENTICATION_SALT))
		return -1; /*already set by busy_handler*/
	recv(rio, *salt, SALTLENGTH, 0);
	
	return 0;
}

uint32_t authenticate(int rio, char * pass)
{
	md5_context ctx;
	unsigned char * md5input;
	unsigned char md5sum[16];

	char * salt;
	uint32_t rights;


	if(get_authentication_salt(rio, &salt) == -1)
		return -1;/* get_authenticatoin_salt had set the errormsg*/

	md5input=(unsigned char *)malloc((strlen(pass)+SALTLENGTH+1));
	memcpy(md5input, salt, SALTLENGTH);
	memcpy(&md5input[SALTLENGTH], pass, strlen(pass));

	md5_starts(&ctx);
	md5_update(&ctx, (uint8 *)md5input, SALTLENGTH+strlen(pass));
	md5_finish(&ctx, md5sum);

	free(salt);
	free(md5input);

	send_request(rio, AUTHENTICATE, (char *)md5sum, 16);
	if(!busy_handler(rio, AUTHENTICATE))
		return -1; /*already set by busy_handler*/
	recv_status(rio); /*we don't care about the status, 
			because we want to get the rights...*/
	recv(rio, &rights, sizeof(rights), 0);
	return rights;
}

int get_device_details(int rio, char **name, char **version, 
			uint32_t * storagedevices)
{
	*name=(char *)malloc(DETAILSLENGTH);
	*version=(char *)malloc(DETAILSLENGTH);

	send_request(rio, GET_DEVICE_DETAILS, NULL, 0);
	if(!busy_handler(rio, GET_DEVICE_DETAILS))
		return -1; /*already set by busy_handler*/

	recv(rio, *name, DETAILSLENGTH, 0);
	recv(rio, *version, DETAILSLENGTH, 0);
	recv(rio, storagedevices, sizeof(uint32_t), 0);
	return 0;
}

int get_storage_details(int rio, uint32_t storage_id, uint32_t *n_files, 
		uint64_t *s_size, uint64_t *f_space, uint32_t * highest_file_id)
{
	char id[3];

	storage_id=htorl(storage_id);
	memcpy(id, &storage_id, sizeof(storage_id));
	send_request(rio, GET_STORAGE_DETAILS, id, sizeof(uint32_t));

	if(!busy_handler(rio, GET_STORAGE_DETAILS)) 
		return -1; /*already set by busy_handler*/
	if(recv_status(rio) != 0){
		lkerrSet(E_FAILEDREQ);
		return -1; /*err: failed request*/
	}

	recv(rio, n_files, sizeof(uint32_t), 0);
	recv(rio, s_size, sizeof(uint64_t), 0);
	recv(rio, f_space, sizeof(uint64_t), 0);
	recv(rio, highest_file_id, sizeof(uint32_t), 0);
	*n_files=rtohl(*n_files);
	*s_size=rtohll(*s_size);
	*f_space=rtohll(*f_space);
	*highest_file_id=rtohl(*highest_file_id);

	return 0;
}

int get_device_settings(int rio, char ** properties)
{
	return recv_properties(rio, GET_DEVICE_SETTINGS, 0, properties);
}


int32_t update_device_settings(int rio, char * properties)
{
	int len;
	len=strlen(properties);
	while(len%4) ++len;
	send_request(rio, UPDATE_DEVICE_SETTINGS, properties, len);
	if(busy_handler(rio, UPDATE_DEVICE_SETTINGS)){
		if(recv_status(rio) != 0){
			lkerrSet(E_FAILEDREQ);
			return -1;/*err: failed request*/
		}else return 0;
	}
	else return -1; /*already set by busy_handler*/
}

int32_t request_io_lock(int rio, uint32_t type)
{
	/* unused: int status; */
	char p[3];

	type=htorl(type);
	memcpy(p, &type, sizeof(type));
	send_request(rio, REQUEST_IO_LOCK, p, 4);
	if(busy_handler(rio, REQUEST_IO_LOCK))
		if(recv_status(rio) != 0){
			lkerrSet(E_FAILEDREQ);
			return -1;/*err: failed request*/
		}else return 0;
	else return -1; /*already set by busy_handler*/
}

int32_t release_io_lock(int rio)
{
	send_request(rio, RELEASE_IO_LOCK, NULL, 0);
	if(busy_handler(rio, RELEASE_IO_LOCK))
		if(recv_status(rio) != 0){
			lkerrSet(E_FAILEDREQ);
			return -1;/*err: failed request*/
		}else return 0;
	else return -1; /*already set by busy_handler*/
}

/*Not yet supported by the Rio Karma.*/
int32_t prepare(int rio, uint64_t size, uint32_t * file_id, uint32_t storage_id)
{
	char p[15];
	
	size=htorl(size);
	*file_id=htorl(*file_id);
	storage_id=htorl(storage_id);
	memcpy(p, &size, sizeof(size));
	memcpy(&(p[4]), file_id, sizeof(file_id));
	memcpy(&(p[12]), &storage_id, sizeof(storage_id));

	send_request(rio, PREPARE, p, 16);
	if(!busy_handler(rio, PREPARE))
		return -1; /*already set by busy_handler*/
	recv(rio, file_id, sizeof(file_id), 0);
	*file_id=rtohl(*file_id);
	
	if(recv_status(rio) != 0){
		lkerrSet(E_FAILEDREQ);
		return -1;/*err: failed request*/
	}else return 0;
}

int32_t write_file_chunk(int rio, uint64_t offset, uint64_t size, 
			 uint32_t file_id, uint32_t storage_id, char * data)
{
	char * packet;
	
	packet=malloc(sizeof(uint64_t)*3+size);
	offset=htorll(offset);
	size=htorll(size);
	file_id=htorl(file_id);
	storage_id=htorl(storage_id);
	memcpy(packet, &offset, sizeof(offset));
	memcpy(&packet[8], &size, sizeof(size));
	size=rtohll(size);
	memcpy(&packet[16], &file_id, sizeof(file_id));
	memcpy(&packet[20], &storage_id, sizeof(storage_id));
	memcpy(&packet[24], data, size);

	send_request(rio, WRITE_FILE_CHUNK, packet, size+sizeof(uint64_t)*3);
	free(packet);
	
	if(busy_handler(rio, WRITE_FILE_CHUNK)){
		if(recv_status(rio) != 0){
			lkerrSet(E_FAILEDREQ);
			return -1;/*err: failed request*/
		}else return 0;
	}
	else return -1; /*already set by busy_handler*/
}

int32_t get_all_file_details(int rio, char ** properties)
{
	return recv_properties(rio, GET_ALL_FILE_DETAILS, 0, properties);
}

int32_t get_file_details(int rio, uint32_t file_id, char ** properties)
{
	return recv_properties(rio, GET_FILE_DETAILS, file_id, properties);
}

int32_t update_file_details(int rio, uint32_t file_id, char * properties)
{
	char * packet; /* unused *tmp; */
	int len,rlen;

	rlen=strlen(properties);
	len=rlen;
	++len;
	while(len%4) ++len;
	packet=malloc(sizeof(uint32_t)+len);
	file_id=htorl(file_id);
	memcpy(packet, &file_id, sizeof(file_id));
	memcpy(&packet[4], properties, len);
	memset(&packet[rlen+sizeof(uint32_t)], 0, len-rlen);
	
	send_request(rio, UPDATE_FILE_DETAILS, packet, len+sizeof(uint32_t));
	free(packet);
	
	if(busy_handler(rio, UPDATE_FILE_DETAILS)){
		inc_devicegeneration();
		if(recv_status(rio) != 0){
			lkerrSet(E_FAILEDREQ);
			return -1;/*err: failed request*/
		}else return 0;
	}else return -1; /*already set by busy_handler*/
}

int32_t read_file_chunk(int rio, uint64_t offset, uint64_t size, 
			uint32_t file_id, char ** data, uint64_t * retsize)
{
	char p[20];
	uint64_t realsize;
	
	*retsize=0;
	offset=htorll(offset);
	size=htorll(size);
	file_id=htorl(file_id);
	memcpy(&p, &offset, sizeof(offset));
	memcpy(&p[8], &size, sizeof(size));
	memcpy(&p[16], &file_id, sizeof(file_id));
	send_request(rio, READ_FILE_CHUNK, p, 20);

	if(!busy_handler(rio, READ_FILE_CHUNK)) 
		return -1; /*already set by busy_handler*/
	recv(rio, &realsize, sizeof(uint64_t), 0);
	realsize=rtohll(realsize);
	*data=(char *)malloc(realsize+1);
	if(recv_status(rio) != 0){
		lkerrSet(E_FAILEDREQ);
		return -1;
	}
	*retsize=recv(rio, *data, realsize, MSG_WAITALL);

	if(recv_status(rio) != 0){
		lkerrSet(E_FAILEDREQ);
		return -1;/*err: failed request*/
	}else return 0;
}
	

int32_t delete_file(int rio,uint32_t file_id)
{
	char id[3];
	
	file_id=htorl(file_id);
	memcpy(id, &file_id, sizeof(file_id));
	send_request(rio, DELETE_FILE, id, 4);
	if(busy_handler(rio, DELETE_FILE)){
		inc_devicegeneration();
		if(recv_status(rio) != 0){
			lkerrSet(E_FAILEDREQ);
			return -1;/*err: failed request*/
		}else return 0;
	}
	else return -1; /*already set by busy_handler*/
}

/* Now, that command would be scary ;)
	 (It's not supported by the rio karma ;))
*/
int32_t format_storage(int rio, uint32_t storage_id)
{
	char id[3];

	storage_id=htorl(storage_id);
	memcpy(id, &storage_id, sizeof(storage_id));
	send_request(rio, FORMAT_STORAGE, id, 4);
	if(busy_handler(rio, FORMAT_STORAGE)){
		if(recv_status(rio) != 0){
			lkerrSet(E_FAILEDREQ);
			return -1;/*err: failed request*/
		}else return 0;
	}
	else return -1; /*already set by busy_handler*/
}

int32_t hangup(int rio)
{
	send_request(rio, HANGUP, NULL, 0);
	if(busy_handler(rio, HANGUP)){
		if(recv_status(rio) != 0){
			lkerrSet(E_FAILEDREQ);
			return -1;/*err: failed request*/
		}else return 0;
	}else return -1; /*already set by busy_handler*/
}

/*Not yet supported by the Rio Karma.*/	
int32_t device_operation(int rio, uint64_t size, char * data, char ** newdata)
{
	char * d;
	
	d=malloc(sizeof(char)*sizeof(uint64_t)+sizeof(char)*size);

	size=htorll(size);
	memcpy(d, &size, sizeof(size));
	size=rtohll(size);
	memcpy(&d[8], data, size);
	send_request(rio, DEVICE_OPERATION, d, size+sizeof(uint64_t));

	if(!busy_handler(rio, DEVICE_OPERATION)) 
		return -1; /*already set by busy_handler*/

	recv(rio, &size, sizeof(uint64_t), 0);
	size=rtohll(size);
	if(recv_status(rio) != 0){
		lkerrSet(E_FAILEDREQ);
		return -1;/*err: failed request*/
	}
	*newdata=realloc(*newdata, size);
	recv(rio, *newdata, size, MSG_WAITALL);
	
	return 0;	
}

/* 
  Advanced Protocol Functions
*/

void update_database(int rio)
{
	char *charset, *settings;
	
	charset=strdup(current_codeset());
	get_device_settings(rio, &settings);
	if(db_changed(settings)){ /*DB was changed*/
		destroy_properties(); 
		init_properties(charset);
		import_properties(settings);
		free(settings);
		get_all_file_details(rio, &settings);
		import_properties(settings);
	}
	free(charset);
	free(settings);
}
