/*
 * libkarma/rio_rw.c
 *
 * Copyright (c) Frank Zschockelt <libkarma@freakysoft.de> 2004
 * 
 * You may distribute and modify this program under the terms of 
 * the GNU GPL, version 2 or later.
 *
 */
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <vorbis/vorbisfile.h>
#include <FLAC/metadata.h>
#include "mp3.h"
#include "wav.h"

#include "karma.h"
#include "properties.h"
#include "util.h"
#include "rio_rw.h"

#define BLOCKSIZE 32768

#define MAXGENRE   147
char *typegenre [MAXGENRE+2] = {
	"Blues","Classic Rock","Country","Dance","Disco","Funk","Grunge",
	"Hip-Hop","Jazz","Metal","New Age","Oldies","Other","Pop","R&B",
	"Rap","Reggae","Rock","Techno","Industrial","Alternative","Ska",
	"Death Metal","Pranks","Soundtrack","Euro-Techno","Ambient",
	"Trip-Hop","Vocal","Jazz+Funk","Fusion","Trance","Classical",
	"Instrumental","Acid","House","Game","Sound Clip","Gospel","Noise",
	"Alt. Rock","Bass","Soul","Punk","Space","Meditative",
	"Instrumental Pop","Instrumental Rock","Ethnic","Gothic",
	"Darkwave","Techno-Industrial","Electronic","Pop-Folk","Eurodance",
	"Dream","Southern Rock","Comedy","Cult","Gangsta Rap","Top 40",
	"Christian Rap","Pop/Funk","Jungle","Native American","Cabaret",
	"New Wave","Psychedelic","Rave","Showtunes","Trailer","Lo-Fi",
	"Tribal","Acid Punk","Acid Jazz","Polka","Retro","Musical",
	"Rock & Roll","Hard Rock","Folk","Folk/Rock","National Folk",
	"Swing","Fast-Fusion","Bebob","Latin","Revival","Celtic",
	"Bluegrass","Avantgarde","Gothic Rock","Progressive Rock",
	"Psychedelic Rock","Symphonic Rock","Slow Rock","Big Band",
	"Chorus","Easy Listening","Acoustic","Humour","Speech","Chanson",
	"Opera","Chamber Music","Sonata","Symphony","Booty Bass","Primus",
	"Porn Groove","Satire","Slow Jam","Club","Tango","Samba",
	"Folklore","Ballad","Power Ballad","Rhythmic Soul","Freestyle",
	"Duet","Punk Rock","Drum Solo","A Cappella","Euro-House",
	"Dance Hall","Goa","Drum & Bass","Club-House","Hardcore","Terror",
	"Indie","BritPop","Negerpunk","Polsk Punk","Beat",
	"Christian Gangsta Rap","Heavy Metal","Black Metal","Crossover",
	"Contemporary Christian","Christian Rock","Merengue","Salsa",
	"Thrash Metal","Anime","JPop","Synthpop",""
};

int synchronize_necessary(int rio)
{
	int ret;
	char *device_generation;
	char *p;

	device_generation=strdup(get_property(0, "device_generation"));
	get_device_settings(rio, &p);
	import_properties(p);
	free(p);
	ret=strcmp(device_generation, get_property(0, "device_generation"));
	free(device_generation);
	return ret;	
}

int rio_read(int rio, uint32_t fid, const char * filename)
{
	int dst;
	uint64_t blk=BLOCKSIZE, offs=0, filesize;
	uint64_t got=0;
	char *tmp;
	
	dst=creat(filename, O_EXCL);
	if(dst==-1) 
		return -1;
	
	filesize=atoll(get_property(fid, "length"));
	
	while(offs<filesize){
		if(read_file_chunk(rio, offs, blk, fid, &tmp, &got) != 0){
			close(dst);
			unlink(filename);
			return -1;
		}
		write(dst, tmp, got);
		free(tmp);
		offs+=got;
		if((filesize-offs) < blk)
			blk=filesize-offs;
	}
	
	close(dst);
	return 0;
}

/*
	 To use that function properly:
	 	1. Get an IO Lock for Writing
	  2. call synchronize_necessary and synchronize the properties if necessary ;)
		3. use rio_write as often as you need to :)
		4. release the IO Lock.
*/
int rio_write(int rio, const char * filename)
{
	int got=-1,fd;
	uint32_t new_fid;
	uint64_t offs=0;
	const uint64_t siz=BLOCKSIZE;
	char * p, tmp[BLOCKSIZE];
	struct stat * len;
	HASH * props;
	
	len=malloc(sizeof(struct stat));
	if(lstat(filename, len)==-1)
		return -1;
	
	new_fid=new_property(); /* uses getnextfid()...*/
	props=idsearch(new_fid);
	if(props==NULL){
		printf("huh, no hash found?");
		return -1;
	}
	p=strrchr(filename, '.');
	
	add_general(props, filename,len->st_size);
	if(p != NULL){
		++p;	
		if(strcmp(p, "ogg") == 0)
			got=get_ogg_props(props, filename);
		else if(strcmp(p, "flac") == 0)
			got=get_flac_props(props, filename, len->st_size);
		else if(strcmp(p, "mp3") == 0)
			got=get_mp3_props(props, filename);
		else if(strcmp(p, "wav") == 0)
			got=get_wave_props(props,filename);
		/*
			 else if(strcmp(p, "wma") == 0)
			 got=get_wma_props(props,filename);
		 */	
		else 
			got=get_taxi_props(props);
	}else got=get_taxi_props(props);

	free(len);
	
	if(got==0){
		fd=open(filename, O_RDONLY);
		while((got=read(fd, &tmp, siz)) > 0){
			if(got < siz) 
				memset(&tmp[got], 0, (siz-got)%4);
			if(write_file_chunk(rio, offs, got+((siz-got)%4), new_fid, 0, tmp)!=0)
			{
				printf("write_file_chunk: %u of %llu\n", got, siz);
				delete_property(new_fid);
				return -1;
			}
			offs+=got+((siz-got)%4);
		}
		close(fd);
		
		p=export_property(new_fid);
		got=update_file_details(rio, new_fid, p);
		free(p);
		return got;
	}else{
		delete_property(new_fid);
		return -1;
	}
}

void add_general(HASH * p, const char * file, unsigned int len)
{
	time_t generation_time;
	char *t;
	
	generation_time=time(NULL);
	change_property_hash(p, "fid_generation", simple_itoa(generation_time));
	change_property_hash(p, "ctime", simple_itoa(generation_time));
	change_property_hash(p, "length", simple_itoa(len));
	t=strrchr(file, '/');
	t=strdup(t?t+1:file);
	change_property_hash(p, "title", t);
	free(t);
}

int get_ogg_props(HASH *p, const char * oggfile)
{
	int rid_fd;
	char bitrate[6];
	char * comment_p;
	FILE * oggf;
	OggVorbis_File ov;
	
	oggf=fopen(oggfile, "r");
	if(ov_open(oggf, &ov, NULL, 0) != 0){
		fclose(oggf);
		return -1;
	}
	
	change_property_hash(p, "codec", "vorbis");
	change_property_hash(p, "type", "tune");
	change_property_hash(p, "samplerate", simple_itoa((ov_info(&ov, -1))->rate) );
	strcpy(bitrate, "vs");
	strcat(bitrate, simple_itoa((ov_info(&ov, -1)->bitrate_nominal)/1000));
	change_property_hash(p, "bitrate", bitrate);
	change_property_hash(p, "duration", simple_itoa(ov_time_total(&ov, -1)*1000));
	
	comment_p=vorbis_comment_query(ov.vc, "title", 0);
	if(comment_p)	change_property_hash(p, "title", comment_p);
	comment_p=vorbis_comment_query(ov.vc, "album", 0);
	if(comment_p) change_property_hash(p, "source", comment_p);
	else change_property_hash(p, "source", NOT_TAGGED);
	comment_p=vorbis_comment_query(ov.vc, "artist", 0);
	if(comment_p) change_property_hash(p, "artist", comment_p);
	else change_property_hash(p, "artist", NOT_TAGGED);
	
	comment_p=vorbis_comment_query(ov.vc, "date", 0);
	if(comment_p) change_property_hash(p, "year", comment_p);
	comment_p=vorbis_comment_query(ov.vc, "tracknumber", 0);
	if(comment_p){
		change_property_hash(p, "tracknr", comment_p);
		change_property_hash(p, "file_id", comment_p);
	}
	comment_p=vorbis_comment_query(ov.vc, "genre", 0);
	if(comment_p) change_property_hash(p, "genre", comment_p);

	comment_p=vorbis_comment_query(ov.vc, "comment", 0);
	if(comment_p) change_property_hash(p, "comment", comment_p);
	
	rid_fd=open(oggfile, O_RDONLY);
	change_property_hash(p, "rid", (char*)generate_rid(rid_fd, 0, ov_raw_total(&ov, 0)));
	close(rid_fd);
	
	ov_clear(&ov);
	
	return 0;
}

int get_flac_props(HASH *p, const char * flacfile, unsigned int len)
{
	int rid_fd;
	FLAC__Metadata_Chain * ch;
	FLAC__Metadata_Iterator * iter;
	FLAC__StreamMetadata * meta;
  	FLAC__StreamMetadata_VorbisComment * vc;
	FLAC__StreamMetadata_VorbisComment_Entry * ve;
	unsigned int c,i,j;
	char *tmp;
	char bitrate[6];
	static const char * tags[] = {"artist", "title", "album", "date", "tracknumber", "genre"};
	
	ch=FLAC__metadata_chain_new();
	if(ch == NULL) return -1;
	if(!FLAC__metadata_chain_read(ch, flacfile)){
		FLAC__metadata_chain_delete(ch);
		return -1;
	}
	/*default values*/
	change_property_hash(p, "source", NOT_TAGGED);
	change_property_hash(p, "artist", NOT_TAGGED);
	
	iter=FLAC__metadata_iterator_new();
	FLAC__metadata_iterator_init(iter, ch);
	do{
		if(FLAC__metadata_iterator_get_block_type(iter)==FLAC__METADATA_TYPE_VORBIS_COMMENT){
			
			meta=FLAC__metadata_iterator_get_block(iter);
			vc=&(meta->data.vorbis_comment);
			
			for(c=0; c < vc->num_comments; c++){
				ve=&(vc->comments[c]);
				
				for(i=0; i<ve->length && ve->entry[i] != '=';i++)
					;
				ve->entry[i]='\0';
				for(j=0; j < sizeof(tags)/sizeof(*tags); ++j){
					if(strcasecmp((char *)ve->entry, tags[j]) == 0){
						if(ve->length-1-i){
							tmp=malloc(ve->length-i+1);
							strncpy(tmp, (char*)&(ve->entry[i+1]), ve->length-1-i);
							tmp[ve->length-i-1]='\0';
							if(j==2) /*album is named "source" on the karma" */
								change_property_hash(p, "source", tmp);
							else if(j==3) /* date -> year */
								change_property_hash(p, "year", tmp);
							else if(j==4){ /*tracknumber->tracknr+file_id*/
								change_property_hash(p, "file_id", tmp);
								change_property_hash(p, "tracknr", tmp);
							}else
								change_property_hash(p, (char*)tags[j], tmp);
							free(tmp);
						}
					}
				}

			}
			
		}else if(FLAC__metadata_iterator_get_block_type(iter)==FLAC__METADATA_TYPE_STREAMINFO){
			meta=FLAC__metadata_iterator_get_block(iter);
			change_property_hash(p, "samplerate", simple_itoa(meta->data.stream_info.sample_rate));
			change_property_hash(p, "duration", simple_itoa(1000 * meta->data.stream_info.total_samples / meta->data.stream_info.sample_rate));
			
			strcpy(bitrate, "vs");
			strcat(bitrate, simple_itoa(8*(len)/(1000*meta->data.stream_info.total_samples/meta->data.stream_info.sample_rate)));
			change_property_hash(p, "bitrate", bitrate);
		}
	}while(FLAC__metadata_iterator_next(iter));

	FLAC__metadata_iterator_delete(iter);
	FLAC__metadata_chain_delete(ch);
	
	change_property_hash(p, "codec", "flac");
	change_property_hash(p, "type", "tune");
	
	rid_fd=open(flacfile, O_RDONLY);
	change_property_hash(p, "rid", (char*)generate_rid(rid_fd, 0, len));
	close(rid_fd);
	
	return 0;
}

int get_mp3_props(HASH *p, const char * mp3file)
{
	mp3info mp3;
	char bitrate[5];
	int rid_fd;
	
	change_property_hash(p, "type", "tune");
	change_property_hash(p, "codec", "mp3");
	/*default values*/
	change_property_hash(p, "artist", NOT_TAGGED);
	change_property_hash(p, "source", NOT_TAGGED);
	
	memset(&mp3,0,sizeof(mp3info));

	mp3.filename=(char *)mp3file;
	mp3.file=fopen(mp3file, "r");
	if(mp3.file==NULL) return -1;
	get_mp3_info(&mp3);
	fclose(mp3.file);
	
	if(mp3.vbr){
		strcpy(bitrate, "vs");
		if(mp3.vbr_average>999)	strcat(bitrate, "999");
		else strcat(bitrate, simple_itoa(mp3.vbr_average));
	}else{
		strcpy(bitrate, "fs");
		strcat(bitrate, simple_itoa(header_bitrate(&mp3.header)));
	}
	
	change_property_hash(p, "bitrate", bitrate);
	change_property_hash(p, "samplerate", simple_itoa(header_frequency(&mp3.header)));
	change_property_hash(p, "duration", simple_itoa(mp3.ms));
	change_property_hash(p, "offset", simple_itoa(mp3.offset));
	
	rid_fd=open(mp3file, O_RDONLY);
	change_property_hash(p, "rid", (char *)generate_rid(rid_fd, mp3.offset, mp3.datasize));
	close(rid_fd);
	
	if(!mp3.id3_isvalid) return 0;
	
	if(mp3.id3.title[0])
		change_property_hash(p, "title", mp3.id3.title);
	if(mp3.id3.artist[0])
		change_property_hash(p, "artist", mp3.id3.artist);
	if(mp3.id3.album[0])
		change_property_hash(p, "source", mp3.id3.album);
	if(mp3.id3.genre[0] <= MAXGENRE)
		change_property_hash(p, "genre", typegenre[(int)mp3.id3.genre[0]]);
	if(mp3.id3.year[0])
		change_property_hash(p, "year", mp3.id3.year);
	if(mp3.id3.comment[0])
		change_property_hash(p, "comment", mp3.id3.comment);
	if(mp3.id3.track[0]){
		change_property_hash(p, "tracknr", simple_itoa(mp3.id3.track[0]));
		change_property_hash(p, "file_id", simple_itoa(mp3.id3.track[0]));
	}
	
	return 0;
}

int get_taxi_props(HASH *p)
{
	change_property_hash(p, "duration", "0");
	change_property_hash(p, "offset", "0");
	change_property_hash(p, "bitrate", "fs128");
	change_property_hash(p, "codec", "taxi");
	change_property_hash(p, "type", "taxi");
	
	return 0;
}
/*
int get_wma_props(HASH *p, const char * wmafile)
{
	return 0;
}
*/
int get_wave_props(HASH *p, const char * wavefile)
{
	int rid_fd;
	wave_header *wh;
	
	if(openwav(&wh, (char*)wavefile) == -1){
		cleanup(&wh);
		return 1;
	}
	change_property_hash(p, "codec", "wave");
	change_property_hash(p, "type", "tune");
	change_property_hash(p, "bitrate", "fs999");
	change_property_hash(p, "source", "wave_file");
	change_property_hash(p, "artist", "wave_file");
	change_property_hash(p, "offset", simple_itoa(sizeof(wave_header)));
	change_property_hash(p, "samplerate", simple_itoa(wh->nSamplesPerSec));
	change_property_hash(p, "duration", 
			simple_itoa(1000 * wh->RiffSize/wh->nChannels/(wh->nChannels * wh->nSamplesPerSec)));
	change_property_hash(p, "file_id", "0");
	change_property_hash(p, "tracknr", "0");
	
	rid_fd=open(wavefile, O_RDONLY);
	change_property_hash(p, "rid", 
			(char*)generate_rid(rid_fd, 0, wh->RiffSize/1024));
	close(rid_fd);

	cleanup(&wh);	
	return 0;
}
