/*
 * libkarma/riocp.c
 *
 * Copyright (c) Frank Zschockelt <libkarma@freakysoft.de> 2004
 *
 * You may distribute and modify this program under the terms of 
 * the GNU GPL, version 2 or later.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <ftw.h>
#include <locale.h>
#include <langinfo.h>

#include "karma.h"
#include "ssdp.h"
#include "properties.h"
#include "rio_rw.h"
#include "errors.h"

#define CHECK(x) x?"failed":"ok"

#define defaultCodeset "ISO8859-15"
#define LOCAL_DB "KARMA_DB"

void usage(void);

void usage(void)
{
	printf("usage: riocp [-a hostname/ip] [-p password] [-c charset]"
		" [files_to_send ...]\n");
}

int main(int argc, char * argv[])
{
	int c, r;
	char *host=NULL;
	char *usercodeset="ISO8859-15";
	char *p=NULL;
	int karma;
	uint16_t port;

	int treecopy(const char *file, const struct stat *sb, int flag)
	{
		if(flag==FTW_F)
			printf("%s: %s\n", file, CHECK(rio_write(karma, file)));
		return 0;
	}

	
	while ((c = getopt(argc, argv, "ha:p:c:")) != -1) {
		switch(c) {
			case 'h':
				usage();
				return -1;
				break;
			case 'a':
				host=strdup(optarg);
				break;
			case 'p':
				p=strdup(optarg);
				break;
			case 'c':
				usercodeset=strdup(optarg);
				break;
			case '?':
				printf("unknown arg %c\n", optopt);
				break;
		}
	}
	lkerrAutoPrint(0);
	if(!host){
		if(ssdp_discover(&host, &port) != 0){
			lkerrorp("*", " -- Can't connect to Rio Karma");
			usage();
			return -1;
		}
		printf("SSDP: %s:%i\n", host, port);
	}
	
	karma=connectkarma(host);
	lkerrorp("","");
	free(host);
	
	if(karma < 0){
		usage();
		if(p) free(p);
		return -1;
	}

	/*init_properties("ISO8859-15");*/
	if(usercodeset != NULL){
		init_properties(usercodeset);
		if (lkerrorp("", "; Trying locale...")) {
			free(usercodeset);
			usercodeset = NULL;
		}
	}
	if(usercodeset == NULL){
		setlocale(LC_CTYPE, "");
		if(strlen(usercodeset = nl_langinfo(CODESET)) == 0) {
			fprintf(stderr, "* Warning: nl_langinfo problem; "
				"using default codeset %s\n", defaultCodeset);
			usercodeset = defaultCodeset;
		}
		init_properties(usercodeset);
		if(! lkerrorp("",""))
			fprintf(stderr, "* Using %s codeset.\n", usercodeset);
	}
	lkerrAutoPrint(1);          /* Setting auto-print-errors ON */
	
	load_properties(LOCAL_DB);

	if(p){
		printf("authenticate: %i\n", authenticate(karma, p));
		free(p);
	}else   printf("authenticate: %i\n", authenticate(karma, ""));
	r = request_io_lock(karma,IO_LOCK_R);
	printf("request_io_lock: %s\n", CHECK(r));

	if(!r) update_database(karma);
	
	if(optind < argc){
		printf("request_io_lock: %s\n", 
			CHECK(request_io_lock(karma,IO_LOCK_W)));
		if(synchronize_necessary(karma)){
			get_all_file_details(karma, &p);
			import_properties(p);
			free(p);
		}
		for(; optind<argc; optind++)
			ftw(argv[optind], treecopy, 1);
		printf("release_io_lock: %s\n", CHECK(release_io_lock(karma))); 
	}
	printf("hangup: %s\n", CHECK(hangup(karma)));
	close(karma);
	save_properties(LOCAL_DB);
	destroy_properties();
	return 0;
}
